﻿'┌────┐
'│技術要綱├──────────────────────
'└┬───┘
'  │『理解不足』
'  │●要るのかどうかわかないけど、《Timer》クラスを解放するために《Dispose》を実装した。
'  │
'  │『マウスカーソルを回避するときの移動について』
'  │・拡大鏡がいつまでもダラダラと動いていると、見難いし見苦しい。
'  │  └→ これを解決するには、高速移動するしかない。
'  │・拡大鏡が瞬間移動すると、目線移動が煩雑になって落ち着かない。しかも、注視している時に突然消えると集中力が途切れてしまう。
'  │  └→ これを解決するには、画面上をスムーズに移動させるしかない。
'  │・終始一定速度で移動する場合にスムーズに見えるようにするには、移動量を小さくしフレームレートを上げなければならない。しかし、フレームレートを上げるにも、《Timer》コントロールの性能的に高が知れている。
'  │  └→ そこで、加速することでこの問題を解決した。初速を遅くし徐々に加速することで、ユーザに「移動することとその方向」を意識させる。脳が移動していることを意識すると、移動量が大きくなって飛び飛びになっても違和感を感じ難くなる（その時、たまたまテレビでっ紹介されていたアニメーションのテクニックを参考にしてみた）。
'  └────────────





''' <summary>
''' マウスカーソルからの逃避行 ～ 基底クラス
''' </summary>
''' <remarks>派生専用。</remarks>
Public MustInherit Class Avoid_Base

    Implements IDisposable



    '===========================================================================================================



    ''' <summary>
    ''' 用意されたマウスカーソルからの回避方法のコレクション
    ''' </summary>
    ''' <remarks>設定コンソールの[ComboBox]用</remarks>
    Public Shared ReadOnly PreparedActions As List(Of PatternItem(Of AvoidActionTypes)) = {New PatternItem(Of AvoidActionTypes) With {.Caption = "手取り足取り", .TypeID = AvoidActionTypes.Passive},
                                                                                           New PatternItem(Of AvoidActionTypes) With {.Caption = "画面中央の右往左往", .TypeID = AvoidActionTypes.ScreenSide},
                                                                                           New PatternItem(Of AvoidActionTypes) With {.Caption = "画面右の三角関係", .TypeID = AvoidActionTypes.RightTriangle},
                                                                                           New PatternItem(Of AvoidActionTypes) With {.Caption = "画面四隅のすれ違い", .TypeID = AvoidActionTypes.ScreenCorner},
                                                                                           New PatternItem(Of AvoidActionTypes) With {.Caption = "画面の中の金魚", .TypeID = AvoidActionTypes.Randezvous}}.ToList



    '===========================================================================================================



    ''' <summary>逃避行動タイミング</summary>
    Public Event Go(ByVal sender As Object, ByVal e As EventArgs)



    '===========================================================================================================



    ''' <summary>初速</summary>
    Protected Const INITIAL_SPEED As Integer = 4

    ''' <summary>加速度</summary>
    Protected Const ACCELERATION As Integer = 4


    '-----------------------------------------------------------------------------------------------------------


    ''' <summary>ビューア</summary>
    Protected Owner As Form


    ''' <summary>逃避行タイミングタイマ</summary>
    Protected WithEvents tmrAvoidTraveler As Timer


    ''' <summary>回避先</summary>
    Protected AvoidDestination As Point

    ''' <summary>回避速度</summary>
    Protected AvoidSpeed As Integer

    ''' <summary>回避発動距離</summary>
    Protected AvoidExerciseDistance As Integer



    '===========================================================================================================



    ''' <summary>
    ''' 回避の初期化（初期位置確定等）
    ''' </summary>
    ''' <param name="WorkArea">対象画面の作業領域</param>
    ''' <remarks>派生クラス側で実装。</remarks>
    Public MustOverride Sub Initialize(WorkArea As Rectangle)

    ''' <summary>
    ''' マウスカーソルからの回避計画を練る
    ''' </summary>
    ''' <param name="WorkArea">対象画面の作業領域</param>
    ''' <remarks>派生クラス側で実装。</remarks>
    Protected MustOverride Sub Plan(WorkArea As Rectangle)



    ':::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::



    ''' <summary>
    ''' インスタンスの初期化
    ''' </summary>
    ''' <param name="Viewer">ビューア</param>
    ''' <remarks></remarks>
    Public Sub New(Viewer As FormMain_Viewer)

        Owner = Viewer

        '=== パラメータの初期化 ===
        Initialize(DirectCast(Owner, FormMain_Viewer).UsingScreen)

        '=== タイマ始動 ===
        tmrAvoidTraveler = New Timer With {.Interval = 33} '約30fps
        tmrAvoidTraveler.Start()

    End Sub

    Private Sub tmrAvoidTraveler_Tick(sender As Object, e As System.EventArgs) Handles tmrAvoidTraveler.Tick

        '=== 回避旅行へ Let's Go! ===
        Plan(DirectCast(Owner, FormMain_Viewer).UsingScreen)
        RaiseEvent Go(Me, New EventArgs)

    End Sub


    '-----------------------------------------------------------------------------------------------------------


    ''' <summary>
    ''' 回避行動
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub Move()

        Dim skip As Point = AvoidDestination - Owner.Location '移動量（目標地点から現在位置を引くことで、その符号から進むべき方向が分かる。）

        If Not skip.IsEmpty Then

            If AvoidSpeed < 1024 Then AvoidSpeed *= ACCELERATION '加速！（1024は念のための規制）

            With skip
                If Math.Abs(.X) > AvoidSpeed Then .X = AvoidSpeed * Math.Sign(.X)
                If Math.Abs(.Y) > AvoidSpeed Then .Y = AvoidSpeed * Math.Sign(.Y)
            End With

            Owner.Location += skip '《Point》を使ってXとYを同時に計算することで、滑らかに移動させる。《Point.Offset(skip)》では、なぜか動かない。

        End If

    End Sub


    '-----------------------------------------------------------------------------------------------------------


#Region "IDisposable Support"
    Private disposedValue As Boolean ' 重複する呼び出しを検出するには

    ' IDisposable
    Protected Overridable Sub Dispose(disposing As Boolean)
        If Not Me.disposedValue Then
            If disposing Then
                ' TODO: マネージ状態を破棄します (マネージ オブジェクト)。
                tmrAvoidTraveler.Dispose()
            End If

            ' TODO: アンマネージ リソース (アンマネージ オブジェクト) を解放し、下の Finalize() をオーバーライドします。
            ' TODO: 大きなフィールドを null に設定します。
        End If
        Me.disposedValue = True
    End Sub

    ' TODO: 上の Dispose(ByVal disposing As Boolean) にアンマネージ リソースを解放するコードがある場合にのみ、Finalize() をオーバーライドします。
    'Protected Overrides Sub Finalize()
    '    ' このコードを変更しないでください。クリーンアップ コードを上の Dispose(ByVal disposing As Boolean) に記述します。
    '    Dispose(False)
    '    MyBase.Finalize()
    'End Sub

    ' このコードは、破棄可能なパターンを正しく実装できるように Visual Basic によって追加されました。
    Public Sub Dispose() Implements IDisposable.Dispose
        ' このコードを変更しないでください。クリーンアップ コードを上の Dispose(ByVal disposing As Boolean) に記述します。
        Dispose(True)
        GC.SuppressFinalize(Me)
    End Sub
#End Region

End Class
